@@ -5,7 +5,7 @@
 %%% Created : 24 Jul 2004 by Alexey Shchepin <alexey@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2017 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2019 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -29,23 +29,22 @@
 
 -behaviour(gen_server).
 
--export([open_nif/8, get_decrypted_input_nif/2,
-	 set_encrypted_input_nif/2, get_encrypted_output_nif/1,
-	 set_decrypted_output_nif/2, get_peer_certificate_nif/1,
-	 get_verify_result_nif/1, invalidate_nif/1, get_negotiated_cipher_nif/1]).
+-export([open_nif/8, loop_nif/4, get_peer_certificate_nif/1,
+         get_verify_result_nif/1, invalidate_nif/1,
+         get_negotiated_cipher_nif/1]).
 
 -export([start_link/0, tcp_to_tls/2,
-	 tls_to_tcp/1, send/2, recv/2, recv/3, recv_data/2,
-	 setopts/2, sockname/1, peername/1,
-	 controlling_process/2, close/1,
-	 get_peer_certificate/1, get_peer_certificate/2,
-	 get_verify_result/1, get_cert_verify_string/2,
-	 add_certfile/2, get_certfile/1, delete_certfile/1,
-	 clear_cache/0, get_negotiated_cipher/1]).
+         tls_to_tcp/1, send/2, recv/2, recv/3, recv_data/2,
+         setopts/2, sockname/1, peername/1,
+         controlling_process/2, close/1,
+         get_peer_certificate/1, get_peer_certificate/2,
+         get_verify_result/1, get_cert_verify_string/2,
+         add_certfile/2, get_certfile/1, delete_certfile/1,
+         clear_cache/0, get_negotiated_cipher/1]).
 
 %% Internal exports, call-back functions.
 -export([init/1, handle_call/3, handle_cast/2,
-	 handle_info/2, code_change/3, terminate/2]).
+         handle_info/2, code_change/3, terminate/2]).
 
 -ifdef(TEST).
 -include_lib("eunit/include/eunit.hrl").
@@ -85,7 +84,7 @@
 
 start_link() ->
     gen_server:start_link({local, ?MODULE}, ?MODULE, [],
-			  []).
+                          []).
 
 init([]) ->
     case load_nif() of
@@ -98,16 +97,7 @@ init([]) ->
 open_nif(_Flags, _CertFile, _Ciphers, _ProtocolOpts, _DHFile, _CAFile, _SNI, _ALPN) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
-get_decrypted_input_nif(_Port, _Length) ->
-    erlang:nif_error({nif_not_loaded, ?MODULE}).
-
-set_encrypted_input_nif(_Port, _Packet) ->
-    erlang:nif_error({nif_not_loaded, ?MODULE}).
-
-get_encrypted_output_nif(_Port) ->
-    erlang:nif_error({nif_not_loaded, ?MODULE}).
-
-set_decrypted_output_nif(_Port, _Packet) ->
+loop_nif(_Port, _ToSend, _Received, _ReceiveBytes) ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
 get_peer_certificate_nif(_Port) ->
@@ -132,7 +122,7 @@ clear_cache_nif() ->
     erlang:nif_error({nif_not_loaded, ?MODULE}).
 
 get_negotiated_cipher_nif(_Port) ->
-	erlang:nif_error({nif_not_loaded, ?MODULE}).
+    erlang:nif_error({nif_not_loaded, ?MODULE}).
 
 %%% --------------------------------------------------------
 %%% The call-back functions.
@@ -154,76 +144,76 @@ terminate(_Reason, _State) ->
     ok.
 
 -spec tcp_to_tls(inet:socket(),
-                 [{atom(), any()}]) -> {'error','no_certfile' | binary()} |
-                                       {ok, tls_socket()}.
+                 [{atom(), any()}]) -> {'error', 'no_certfile' | binary()} |
+{ok, tls_socket()}.
 
 tcp_to_tls(TCPSocket, Options) ->
     Command = case lists:member(connect, Options) of
-		  true -> ?SET_CERTIFICATE_FILE_CONNECT;
-		  false -> ?SET_CERTIFICATE_FILE_ACCEPT
-	      end,
+                  true -> ?SET_CERTIFICATE_FILE_CONNECT;
+                  false -> ?SET_CERTIFICATE_FILE_ACCEPT
+              end,
     CertFile = proplists:get_value(certfile, Options, ""),
     if CertFile /= [] orelse Command == ?SET_CERTIFICATE_FILE_ACCEPT ->
-	    Flags1 = case lists:member(verify_none, Options) of
-			 true -> ?VERIFY_NONE;
-			 false -> 0
-		     end,
-	    Flags2 = case lists:member(compression_none, Options) of
-			 true -> ?COMPRESSION_NONE;
-			 false -> 0
-		     end,
-	    Flags = Flags1 bor Flags2,
-	    Ciphers =
-	    case lists:keysearch(ciphers, 1, Options) of
-		{value, {ciphers, C}} ->
-		    iolist_to_binary(C);
-		false ->
-		    <<>>
-	    end,
-	    ProtocolOpts = case lists:keysearch(protocol_options, 1, Options) of
-			       {value, {protocol_options, P}} ->
-				   iolist_to_binary(P);
-			       false ->
-				   <<>>
-			   end,
-	    DHFile = case lists:keysearch(dhfile, 1, Options) of
-			 {value, {dhfile, D}} ->
-			     iolist_to_binary(D);
-			 false ->
-			     <<>>
-		     end,
-	    CAFile = case lists:keysearch(cafile, 1, Options) of
-			 {value, {cafile, CA}} ->
-			     iolist_to_binary(CA);
-			 false ->
-			     <<>>
-		     end,
-	    ServerName = case lists:keysearch(sni, 1, Options) of
-			     {value, {sni, SNI}} ->
-				 iolist_to_binary(SNI);
-			     false ->
-				 <<>>
-			 end,
-	    ALPN = case lists:keysearch(alpn, 1, Options) of
-		       {value, {alpn, ProtoList}} ->
-			   encode_alpn(ProtoList);
-		       false ->
-			   <<>>
-		   end,
-	    case open_nif(Command bor Flags, CertFile, Ciphers, ProtocolOpts,
-			  DHFile, CAFile, ServerName, ALPN) of
-		{ok, Port} ->
-		    {ok, #tlssock{tcpsock = TCPSocket, tlsport = Port}};
-		Err = {error, _} ->
-		    Err
-	    end;
-	true -> {error, no_certfile}
+        Flags1 = case lists:member(verify_none, Options) of
+                     true -> ?VERIFY_NONE;
+                     false -> 0
+                 end,
+        Flags2 = case lists:member(compression_none, Options) of
+                     true -> ?COMPRESSION_NONE;
+                     false -> 0
+                 end,
+        Flags = Flags1 bor Flags2,
+        Ciphers =
+        case lists:keysearch(ciphers, 1, Options) of
+            {value, {ciphers, C}} ->
+                iolist_to_binary(C);
+            false ->
+                <<>>
+        end,
+        ProtocolOpts = case lists:keysearch(protocol_options, 1, Options) of
+                           {value, {protocol_options, P}} ->
+                               iolist_to_binary(P);
+                           false ->
+                               <<>>
+                       end,
+        DHFile = case lists:keysearch(dhfile, 1, Options) of
+                     {value, {dhfile, D}} ->
+                         iolist_to_binary(D);
+                     false ->
+                         <<>>
+                 end,
+        CAFile = case lists:keysearch(cafile, 1, Options) of
+                     {value, {cafile, CA}} ->
+                         iolist_to_binary(CA);
+                     false ->
+                         <<>>
+                 end,
+        ServerName = case lists:keysearch(sni, 1, Options) of
+                         {value, {sni, SNI}} ->
+                             iolist_to_binary(SNI);
+                         false ->
+                             <<>>
+                     end,
+        ALPN = case lists:keysearch(alpn, 1, Options) of
+                   {value, {alpn, ProtoList}} ->
+                       encode_alpn(ProtoList);
+                   false ->
+                       <<>>
+               end,
+        case open_nif(Command bor Flags, CertFile, Ciphers, ProtocolOpts,
+                      DHFile, CAFile, ServerName, ALPN) of
+            {ok, Port} ->
+                {ok, #tlssock{tcpsock = TCPSocket, tlsport = Port}};
+            Err = {error, _} ->
+                Err
+        end;
+        true -> {error, no_certfile}
     end.
 
 -spec tls_to_tcp(tls_socket()) -> inet:socket().
 
 tls_to_tcp(#tlssock{tcpsock = TCPSocket,
-		    tlsport = Port}) ->
+                    tlsport = Port}) ->
     invalidate_nif(Port),
     TCPSocket.
 
@@ -231,86 +221,82 @@ recv(Socket, Length) -> recv(Socket, Length, infinity).
 
 -spec recv(tls_socket(), non_neg_integer(),
            timeout()) -> {error, inet:posix()} |
-                         {error, binary()} |
-                         {ok, binary()}.
+{error, binary()} |
+{ok, binary()}.
 
-recv(#tlssock{tcpsock = TCPSocket} =
-	 TLSSock,
-     Length, Timeout) ->
-    case recv_data(TLSSock, <<>>, Length) of
-        {ok, <<>>} ->
-            case gen_tcp:recv(TCPSocket, 0, Timeout) of
-                {ok, Packet} -> recv_data(TLSSock, Packet, Length);
-                {error, _Reason} = Error -> Error
-            end;
-        Res -> Res
-    end.
+recv(TLSSock, Length, Timeout) ->
+    recv_and_loop(TLSSock, <<>>, <<>>, <<>>,
+                  case Length of 0 -> -1; _ -> Length end, Timeout).
 
 -spec recv_data(tls_socket(), binary()) -> {error, inet:posix() | binary()} |
-                                           {ok, binary()}.
+{ok, binary()}.
 
 recv_data(TLSSock, Packet) ->
-    recv_data(TLSSock, Packet, 0).
+    loop(TLSSock, <<>>, Packet, <<>>, -1).
 
--spec recv_data(tls_socket(), binary(),
-                non_neg_integer()) -> {error, inet:posix() | binary()} |
-                                      {ok, binary()}.
-
-recv_data(TLSSock, Packet, Length) ->
-    case catch recv_data1(TLSSock, Packet, Length) of
-      {'EXIT', Reason} -> {error, Reason};
-      Res -> Res
+-spec loop(tls_socket(), binary(), binary(), binary(), integer()) ->
+    {error, inet:posix() | binary()} |
+    {ok, binary()}.
+loop(#tlssock{tcpsock = TCPSocket,
+              tlsport = Port} = Socket,
+     ToSend, Received, DecBuf, Length) ->
+    try loop_nif(Port, ToSend, Received, Length) of
+        {error, _} = Err ->
+            Err;
+        {<<>>, Decrypted} ->
+            {ok, <<DecBuf/binary, Decrypted/binary>>};
+        {ToWrite, Decrypted} ->
+            case gen_tcp:send(TCPSocket, ToWrite) of
+                ok ->
+                    loop(Socket, <<>>, <<>>, <<DecBuf/binary, Decrypted/binary>>,
+                         Length - byte_size(Decrypted));
+                {error, _} = Err ->
+                    Err
+            end
+    catch error:badarg ->
+        {error, einval}
     end.
 
-recv_data1(#tlssock{tcpsock = TCPSocket,
-		    tlsport = Port},
-	   Packet, Length) ->
-    case catch set_encrypted_input_nif(Port, Packet) of
-	{'EXIT', {badarg, _}} ->
-	    {error, einval};
-	ok ->
-	    case catch get_decrypted_input_nif(Port, Length) of
-		{'EXIT', {badarg, _}} ->
-		    {error, einval};
-		{ok, In} -> {ok, In};
-		{send, In} ->
-		    case catch get_encrypted_output_nif(Port) of
-			{'EXIT', {badarg, _}} ->
-			    {error, einval};
-			{ok, Out} ->
-			    case gen_tcp:send(TCPSocket, Out) of
-				ok -> {ok, In};
-				Error -> Error
-			    end;
-			{error, _} = Err ->
-			    Err
-		    end;
-		{error, _} = Err ->
-		    Err
-	    end;
-	{error, _} = Err ->
-	    Err
+-spec recv_and_loop(tls_socket(), binary(), binary(), binary(), integer(), integer()) ->
+    {error, inet:posix() | binary()} |
+    {ok, binary()}.
+recv_and_loop(#tlssock{tcpsock = TCPSocket} = Socket,
+              ToSend, Received, DecBuf, Length, Timeout) ->
+    case loop(Socket, ToSend, Received, DecBuf, Length) of
+        {error, _} = Err ->
+            Err;
+        {ok, Decrypted} ->
+            case size(Decrypted) - size(DecBuf) of
+                V when V == 0 ->
+                    case gen_tcp:recv(TCPSocket, 0, Timeout) of
+                        {ok, Received2} ->
+                            recv_and_loop(Socket, <<>>, Received2, Decrypted,
+                                          Length - V, Timeout);
+                        {error, _} = Err ->
+                            Err
+                    end;
+                V when Length > V ->
+                    recv_and_loop(Socket, <<>>, <<>>, Decrypted,
+                                  Length - V, Timeout);
+                _ when Length < 0 ->
+                    {ok, Decrypted};
+                V when Length == V ->
+                    {ok, Decrypted};
+                _ ->
+                    {error, too_much_data_received}
+            end
     end.
 
--spec send(tls_socket(), binary()) -> ok | {error, inet:posix() |
-                                            binary() | timeout}.
+-spec send(tls_socket(), binary()) -> ok | {error, inet:posix() | binary() | timeout}.
 
-send(#tlssock{tcpsock = TCPSocket, tlsport = Port},
-     Packet) ->
-    case catch set_decrypted_output_nif(Port, Packet) of
-	{'EXIT', {badarg, _}} ->
-	    {error, einval};
-	ok ->
-	    case catch get_encrypted_output_nif(Port) of
-		{'EXIT', {badarg, _}} ->
-		    {error, einval};
-		{ok, Out} ->
-		    gen_tcp:send(TCPSocket, Out);
-		{error, _} = Err ->
-		    Err
-	    end;
-	{error, _} = Err ->
-	    Err
+send(Socket, Packet) ->
+    case loop(Socket, Packet, <<>>, <<>>, 0) of
+        {ok, <<>>} ->
+            ok;
+        {ok, Data} ->
+            Data;
+        Other ->
+            Other
     end.
 
 -spec setopts(tls_socket(), list()) -> ok | {error, inet:posix()}.
@@ -319,7 +305,7 @@ setopts(#tlssock{tcpsock = TCPSocket}, Opts) ->
     inet:setopts(TCPSocket, Opts).
 
 -spec sockname(tls_socket()) -> {ok, {inet:ip_address(), inet:port_number()}} |
-                                {error, inet:posix()}.
+{error, inet:posix()}.
 
 sockname(#tlssock{tcpsock = TCPSocket}) ->
     inet:sockname(TCPSocket).
@@ -328,7 +314,7 @@ peername(#tlssock{tcpsock = TCPSocket}) ->
     inet:peername(TCPSocket).
 
 controlling_process(#tlssock{tcpsock = TCPSocket},
-		    Pid) ->
+                    Pid) ->
     gen_tcp:controlling_process(TCPSocket, Pid).
 
 close(#tlssock{tcpsock = TCPSocket, tlsport = Port}) ->
@@ -342,25 +328,25 @@ get_peer_certificate(TLSSock) ->
 -spec get_peer_certificate(tls_socket(), otp|plain) -> error | {ok, cert()}.
 get_peer_certificate(#tlssock{tlsport = Port}, Type) ->
     case catch get_peer_certificate_nif(Port) of
-	{'EXIT', {badarg, _}} ->
-	    error;
-	{ok, BCert} ->
-	    try public_key:pkix_decode_cert(BCert, Type) of
-		Cert -> {ok, Cert}
-	    catch _:_ ->
-		    error
-	    end;
-	{error, _} -> error
+        {'EXIT', {badarg, _}} ->
+            error;
+        {ok, BCert} ->
+            try public_key:pkix_decode_cert(BCert, Type) of
+                Cert -> {ok, Cert}
+            catch _:_ ->
+                error
+            end;
+        {error, _} -> error
     end.
 
 -spec get_negotiated_cipher(tls_socket()) -> error | {ok, binary()}.
 get_negotiated_cipher(#tlssock{tlsport = Port}) ->
-		case catch get_negotiated_cipher_nif(Port) of
-			Val when is_binary(Val) ->
-				{ok, Val};
-			_ ->
-				error
-		end.
+    case catch get_negotiated_cipher_nif(Port) of
+        Val when is_binary(Val) ->
+            {ok, Val};
+        _ ->
+            error
+    end.
 
 -spec get_verify_result(tls_socket()) -> byte().
 
@@ -372,12 +358,12 @@ get_verify_result(#tlssock{tlsport = Port}) ->
 
 get_cert_verify_string(CertVerifyRes, Cert) ->
     case catch cert_is_self_signed(Cert) of
-      {'EXIT', _} -> <<"unknown verification error">>;
-      IsSelfsigned ->
-	  case {CertVerifyRes, IsSelfsigned} of
-	    {21, true} -> <<"self-signed certificate">>;
-	    _ -> cert_verify_code(CertVerifyRes)
-	  end
+        {'EXIT', _} -> <<"unknown verification error">>;
+        IsSelfsigned ->
+            case {CertVerifyRes, IsSelfsigned} of
+                {21, true} -> <<"self-signed certificate">>;
+                _ -> cert_verify_code(CertVerifyRes)
+            end
     end.
 
 cert_is_self_signed(#'Certificate'{} = Cert) ->
@@ -407,7 +393,7 @@ delete_certfile(Domain) ->
 clear_cache() ->
     clear_cache_nif().
 
-cert_verify_code(0) -> <<"ok">>;
+cert_verify_code(0)  -> <<"ok">>;
 cert_verify_code(2) ->
     <<"unable to get issuer certificate">>;
 cert_verify_code(3) ->
@@ -420,7 +406,7 @@ cert_verify_code(6) ->
     <<"unable to decode issuer public key">>;
 cert_verify_code(7) ->
     <<"certificate signature failure">>;
-cert_verify_code(8) -> <<"CRL signature failure">>;
+cert_verify_code(8)  -> <<"CRL signature failure">>;
 cert_verify_code(9) ->
     <<"certificate is not yet valid">>;
 cert_verify_code(10) -> <<"certificate has expired">>;
@@ -502,13 +488,13 @@ transmission_test() ->
     SPid = setup_sender(Port, []),
     SPid ! {stop, self()},
     receive
-	{result, Res} ->
-	    ?assertEqual(ok, Res)
+        {result, Res} ->
+            ?assertEqual(ok, Res)
     end,
     LPid ! {stop, self()},
     receive
-	{received, Msg} ->
-	    ?assertEqual(<<"abcdefghi">>, Msg)
+        {received, Msg} ->
+            ?assertEqual(<<"abcdefghi">>, Msg)
     end.
 
 not_compatible_protocol_options_test() ->
@@ -516,80 +502,80 @@ not_compatible_protocol_options_test() ->
     SPid = setup_sender(Port, [{protocol_options, <<"no_sslv2|no_sslv3|no_tlsv1|no_tlsv1_2|no_tlsv1_3">>}]),
     SPid ! {stop, self()},
     receive
-	{result, Res} ->
-	    ?assertMatch({badmatch, {error, _}}, Res)
+        {result, Res} ->
+            ?assertMatch({badmatch, {error, _}}, Res)
     end,
     LPid ! {stop, self()},
     receive
-	{received, {error, _, _} = Msg} ->
-	    ?assertMatch({error, _, <<>>}, Msg);
-	{received, Msg} ->
-	    ?assertMatch(<<>>, Msg)
+        {received, {error, _, _} = Msg} ->
+            ?assertMatch({error, _, <<>>}, Msg);
+        {received, Msg} ->
+            ?assertMatch(<<>>, Msg)
     end.
 
 setup_listener(Opts) ->
     {ok, ListenSocket} = gen_tcp:listen(0,
-					[binary, {packet, 0}, {active, false},
-					 {reuseaddr, true}, {nodelay, true}]),
+                                        [binary, {packet, 0}, {active, false},
+                                         {reuseaddr, true}, {nodelay, true}]),
     Pid = spawn(fun() ->
-	{ok, Socket} = gen_tcp:accept(ListenSocket),
-	{ok, TLSSock} = tcp_to_tls(Socket, [{certfile, <<"../tests/cert.pem">>} | Opts]),
-	listener_loop(TLSSock, <<>>)
-		end),
+        {ok, Socket} = gen_tcp:accept(ListenSocket),
+        {ok, TLSSock} = tcp_to_tls(Socket, [{certfile, <<"../tests/cert.pem">>} | Opts]),
+        listener_loop(TLSSock, <<>>)
+                end),
     {ok, Port} = inet:port(ListenSocket),
     {Pid, Port}.
 
 listener_loop(TLSSock, Msg) ->
     case recv(TLSSock, 1, 1000) of
-	{error, timeout} ->
-	    receive
-		{stop, Pid} ->
-		    Pid ! {received, Msg}
-	    after 0 ->
-		listener_loop(TLSSock, Msg)
-	    end;
-	{error, closed} ->
-	    receive
-		{stop, Pid} ->
-		    Pid ! {received, Msg}
-	    end;
-	{error, Err} ->
-	    receive
-		{stop, Pid} ->
-		    Pid ! {received, {error, Err, Msg}}
-	    end;
-	{ok, Data} ->
-	    listener_loop(TLSSock, <<Msg/binary, Data/binary>>)
+        {error, timeout} ->
+            receive
+                {stop, Pid} ->
+                    Pid ! {received, Msg}
+            after 0 ->
+                listener_loop(TLSSock, Msg)
+            end;
+        {error, closed} ->
+            receive
+                {stop, Pid} ->
+                    Pid ! {received, Msg}
+            end;
+        {error, Err} ->
+            receive
+                {stop, Pid} ->
+                    Pid ! {received, {error, Err, Msg}}
+            end;
+        {ok, Data} ->
+            listener_loop(TLSSock, <<Msg/binary, Data/binary>>)
     end.
 
 setup_sender(Port, Opts) ->
     {ok, Socket} = gen_tcp:connect({127, 0, 0, 1}, Port, [
-	binary, {packet, 0}, {active, false},
-	{reuseaddr, true}, {nodelay, true}]),
+        binary, {packet, 0}, {active, false},
+        {reuseaddr, true}, {nodelay, true}]),
     spawn(fun() ->
-	{ok, TLSSock} = tcp_to_tls(Socket, [connect, {certfile, <<"../tests/cert.pem">>} | Opts]),
-	sender_loop(TLSSock)
-	  end).
+        {ok, TLSSock} = tcp_to_tls(Socket, [connect, {certfile, <<"../tests/cert.pem">>} | Opts]),
+        sender_loop(TLSSock)
+          end).
 
 sender_loop(TLSSock) ->
     Res = try
-	      recv(TLSSock, 0, 1000),
-	      ok = send(TLSSock, <<"abc">>),
-	      recv(TLSSock, 0, 1000),
-	      ok = send(TLSSock, <<"def">>),
-	      recv(TLSSock, 0, 1000),
-	      ok = send(TLSSock, <<"ghi">>),
-	      recv(TLSSock, 0, 1000),
-	      close(TLSSock),
-	      ok
-	  catch
-	      _:Err ->
-	      close(TLSSock),
-	      Err
-	  end,
+              recv(TLSSock, 0, 100),
+              ok = send(TLSSock, <<"abc">>),
+              recv(TLSSock, 0, 100),
+              ok = send(TLSSock, <<"def">>),
+              recv(TLSSock, 0, 100),
+              ok = send(TLSSock, <<"ghi">>),
+              recv(TLSSock, 0, 100),
+              close(TLSSock),
+              ok
+          catch
+              _:Err ->
+                  close(TLSSock),
+                  Err
+          end,
     receive
-	{stop, Pid} ->
-	    Pid ! {result, Res}
+        {stop, Pid} ->
+            Pid ! {result, Res}
     end.
 
 -endif.
